#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "verify_host_resolution",
    "create_swagger_client",
    "verify_client_version",
    "update_session_proxies",
    "build_operation_url",
    "handle_server_raw_response_messages",
    "NeptuneResponseAdapter",
    "MissingApiClient",
    "cache",
    "ssl_verify",
    "parse_validation_errors",
    "ExecuteOperationsBatchingManager",
    "which_progress_bar",
    "construct_progress_bar",
    "with_api_exceptions_handler",
]

import dataclasses
import itertools
import os
import socket
import time
from functools import (
    lru_cache,
    wraps,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Text,
    Type,
)
from urllib.parse import (
    urljoin,
    urlparse,
)

import requests
import urllib3
from bravado.client import SwaggerClient
from bravado.exception import (
    BravadoConnectionError,
    BravadoTimeoutError,
    HTTPBadGateway,
    HTTPClientError,
    HTTPError,
    HTTPForbidden,
    HTTPGatewayTimeout,
    HTTPInternalServerError,
    HTTPRequestTimeout,
    HTTPServiceUnavailable,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from bravado.http_client import HttpClient
from bravado.requests_client import RequestsResponseAdapter
from bravado_core.formatter import SwaggerFormat
from bravado_core.util import RecursiveCallException
from packaging.version import Version
from requests import (
    Response,
    Session,
)
from requests.exceptions import ChunkedEncodingError
from urllib3.exceptions import NewConnectionError

from neptune.envs import NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE
from neptune.exceptions import (
    CannotResolveHostname,
    MetadataInconsistency,
    NeptuneClientUpgradeRequiredError,
    NeptuneFeatureNotAvailableException,
)
from neptune.internal.backends.api_model import ClientConfig
from neptune.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
from neptune.internal.envs import NEPTUNE_RETRIES_TIMEOUT_ENV
from neptune.internal.exceptions import (
    ClientHttpError,
    Forbidden,
    NeptuneAuthTokenExpired,
    NeptuneConnectionLostException,
    NeptuneInvalidApiTokenException,
    NeptuneSSLVerificationError,
    Unauthorized,
)
from neptune.internal.operation import (
    CopyAttribute,
    Operation,
)
from neptune.internal.utils import replace_patch_version
from neptune.internal.utils.logger import get_logger
from neptune.internal.utils.utils import reset_internal_ssl_state
from neptune.internal.warnings import (
    NeptuneWarning,
    warn_once,
)
from neptune.typing import (
    ProgressBarCallback,
    ProgressBarType,
)
from neptune.utils import (
    NullProgressBar,
    TqdmProgressBar,
)

logger = get_logger()

if TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend


MAX_RETRY_TIME = 30
MAX_RETRY_MULTIPLIER = 10
retries_timeout = int(os.getenv(NEPTUNE_RETRIES_TIMEOUT_ENV, "60"))


def get_retry_from_headers_or_default(headers, retry_count):
    try:
        return (
            int(headers["retry-after"][0]) if "retry-after" in headers else 2 ** min(MAX_RETRY_MULTIPLIER, retry_count)
        )
    except Exception:
        return min(2 ** min(MAX_RETRY_MULTIPLIER, retry_count), MAX_RETRY_TIME)


def with_api_exceptions_handler(func):
    def wrapper(*args, **kwargs):
        ssl_error_occurred = False
        last_exception = None
        start_time = time.monotonic()
        for retry in itertools.count(0):
            if time.monotonic() - start_time > retries_timeout:
                break

            try:
                return func(*args, **kwargs)
            except requests.exceptions.InvalidHeader as e:
                if "X-Neptune-Api-Token" in e.args[0]:
                    raise NeptuneInvalidApiTokenException()
                raise
            except requests.exceptions.SSLError as e:
                """
                OpenSSL's internal random number generator does not properly handle forked processes.
                Applications must change the PRNG state of the parent process
                if they use any SSL feature with os.fork().
                Any successful call of RAND_add(), RAND_bytes() or RAND_pseudo_bytes() is sufficient.
                https://docs.python.org/3/library/ssl.html#multi-processing
                On Linux it looks like it does not help much but does not break anything either.
                But single retry seems to solve the issue.
                """
                if not ssl_error_occurred:
                    ssl_error_occurred = True
                    reset_internal_ssl_state()
                    continue

                if "CertificateError" in str(e.__context__):
                    raise NeptuneSSLVerificationError() from e
                else:
                    time.sleep(min(2 ** min(MAX_RETRY_MULTIPLIER, retry), MAX_RETRY_TIME))
                    last_exception = e
                    continue
            except (
                BravadoConnectionError,
                BravadoTimeoutError,
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                HTTPRequestTimeout,
                HTTPServiceUnavailable,
                HTTPGatewayTimeout,
                HTTPBadGateway,
                HTTPInternalServerError,
                NewConnectionError,
                ChunkedEncodingError,
                RecursiveCallException,
            ) as e:
                time.sleep(min(2 ** min(MAX_RETRY_MULTIPLIER, retry), MAX_RETRY_TIME))
                last_exception = e
                continue
            except HTTPTooManyRequests as e:
                wait_time = get_retry_from_headers_or_default(e.response.headers, retry)
                time.sleep(wait_time)
                last_exception = e
                continue
            except NeptuneAuthTokenExpired as e:
                last_exception = e
                continue
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except HTTPClientError as e:
                raise ClientHttpError(e.status_code, e.response.text) from e
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code in (
                    HTTPRequestTimeout.status_code,
                    HTTPBadGateway.status_code,
                    HTTPServiceUnavailable.status_code,
                    HTTPGatewayTimeout.status_code,
                    HTTPInternalServerError.status_code,
                ):
                    time.sleep(min(2 ** min(MAX_RETRY_MULTIPLIER, retry), MAX_RETRY_TIME))
                    last_exception = e
                    continue
                elif status_code == HTTPTooManyRequests.status_code:
                    wait_time = get_retry_from_headers_or_default(e.response.headers, retry)
                    time.sleep(wait_time)
                    last_exception = e
                    continue
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                elif 400 <= status_code < 500:
                    raise ClientHttpError(status_code, e.response.text) from e
                else:
                    raise
        raise NeptuneConnectionLostException(last_exception) from last_exception

    return wrapper


@lru_cache(maxsize=None, typed=True)
def verify_host_resolution(url: str) -> None:
    host = urlparse(url).netloc.split(":")[0]
    try:
        socket.gethostbyname(host)
    except socket.gaierror:
        raise CannotResolveHostname(host)


uuid_format = SwaggerFormat(
    format="uuid",
    to_python=lambda x: x,
    to_wire=lambda x: x,
    validate=lambda x: None,
    description="",
)


@with_api_exceptions_handler
def create_swagger_client(url: str, http_client: HttpClient) -> SwaggerClient:
    return SwaggerClient.from_url(
        url,
        config=dict(
            validate_swagger_spec=False,
            validate_requests=False,
            validate_responses=False,
            formats=[uuid_format],
        ),
        http_client=http_client,
    )


def verify_client_version(client_config: ClientConfig, version: Version):
    version_with_patch_0 = Version(replace_patch_version(str(version)))
    if client_config.version_info.min_compatible and client_config.version_info.min_compatible > version:
        raise NeptuneClientUpgradeRequiredError(version, min_version=client_config.version_info.min_compatible)
    if client_config.version_info.max_compatible and client_config.version_info.max_compatible < version_with_patch_0:
        raise NeptuneClientUpgradeRequiredError(version, max_version=client_config.version_info.max_compatible)
    if client_config.version_info.min_recommended and client_config.version_info.min_recommended > version:
        logger.warning(
            "WARNING: Your version of the Neptune client library (%s) is deprecated,"
            " and soon will no longer be supported by the Neptune server."
            " We recommend upgrading to at least version %s.",
            version,
            client_config.version_info.min_recommended,
        )


def update_session_proxies(session: Session, proxies: Optional[Dict[str, str]]):
    if proxies:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError(f"Wrong proxies format: {proxies}")


def build_operation_url(base_api: str, operation_url: str) -> str:
    if "://" not in base_api:
        base_api = f"https://{base_api}"

    return urljoin(base=base_api, url=operation_url)


# TODO print in color once colored exceptions are added
def handle_server_raw_response_messages(response: Response):
    try:
        info = response.headers.get("X-Server-Info")
        if info:
            logger.info(info)
        warning = response.headers.get("X-Server-Warning")
        if warning:
            logger.warning(warning)
        error = response.headers.get("X-Server-Error")
        if error:
            logger.error(error)
        return response
    except Exception:
        # any issues with printing server messages should not cause code to fail
        return response


# TODO print in color once colored exceptions are added
class NeptuneResponseAdapter(RequestsResponseAdapter):
    @property
    def raw_bytes(self) -> bytes:
        self._handle_response()
        return super().raw_bytes

    @property
    def text(self) -> Text:
        self._handle_response()
        return super().text

    def json(self, **kwargs) -> Mapping[Text, Any]:
        self._handle_response()
        return super().json(**kwargs)

    def _handle_response(self):
        try:
            info = self._delegate.headers.get("X-Server-Info")
            if info:
                logger.info(info)
            warning = self._delegate.headers.get("X-Server-Warning")
            if warning:
                logger.warning(warning)
            error = self._delegate.headers.get("X-Server-Error")
            if error:
                logger.error(error)
        except Exception:
            # any issues with printing server messages should not cause code to fail
            pass


class MissingApiClient(SwaggerClientWrapper):
    """catch-all class to gracefully handle calls to unavailable API"""

    def __init__(self, feature_name: str):
        self.feature_name = feature_name

    def __getattr__(self, item):
        raise NeptuneFeatureNotAvailableException(missing_feature=self.feature_name)


# https://stackoverflow.com/a/44776960
def cache(func):
    """
    Transform mutable dictionary into immutable before call to lru_cache
    """

    class HDict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

    func = lru_cache(maxsize=None, typed=True)(func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        args = tuple([HDict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: HDict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)

    wrapper.cache_clear = func.cache_clear
    return wrapper


def ssl_verify():
    if os.getenv(NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE):
        urllib3.disable_warnings()
        return False

    return True


def parse_validation_errors(error: HTTPError) -> Dict[str, str]:
    return {
        f"{error_description.get('errorCode').get('name')}": error_description.get("context", "")
        for validation_error in error.swagger_result.validationErrors
        for error_description in validation_error.get("errors")
    }


@dataclasses.dataclass
class OperationsBatch:
    operations: List[Operation] = dataclasses.field(default_factory=list)
    errors: List[MetadataInconsistency] = dataclasses.field(default_factory=list)
    dropped_operations_count: int = 0


class ExecuteOperationsBatchingManager:
    def __init__(self, backend: "NeptuneBackend"):
        self._backend = backend

    def get_batch(self, ops: Iterable[Operation]) -> OperationsBatch:
        result = OperationsBatch()
        for op in ops:
            if isinstance(op, CopyAttribute):
                if not result.operations:
                    try:
                        # CopyAttribute can be at the start of a batch
                        result.operations.append(op.resolve(self._backend))
                    except MetadataInconsistency as e:
                        result.errors.append(e)
                        result.dropped_operations_count += 1
                else:
                    # cannot have CopyAttribute after any other op in a batch
                    break
            else:
                result.operations.append(op)

        return result


def _check_if_tqdm_installed() -> bool:
    try:
        import tqdm  # noqa: F401

        return True
    except ImportError:  # tqdm not installed
        return False


def which_progress_bar(progress_bar: Optional[ProgressBarType]) -> Type[ProgressBarCallback]:
    if isinstance(progress_bar, type) and issubclass(
        progress_bar, ProgressBarCallback
    ):  # return whatever the user gave us
        return progress_bar

    if not isinstance(progress_bar, bool) and progress_bar is not None:
        raise TypeError(f"progress_bar should be None, bool or ProgressBarCallback, got {type(progress_bar).__name__}")

    if progress_bar or progress_bar is None:
        tqdm_available = _check_if_tqdm_installed()

        if not tqdm_available:
            warn_once(
                "To use the default progress bar, please install tqdm: pip install tqdm",
                exception=NeptuneWarning,
            )
            return NullProgressBar
        return TqdmProgressBar

    return NullProgressBar


def construct_progress_bar(
    progress_bar: Optional[ProgressBarType],
    description: str,
) -> ProgressBarCallback:
    progress_bar_type = which_progress_bar(progress_bar)
    return progress_bar_type(description=description)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = ["to_pandas"]

from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
    Tuple,
    Union,
)

import pandas as pd

from neptune.api.models import (
    ArtifactField,
    BoolField,
    DateTimeField,
    FieldVisitor,
    FileField,
    FileSetField,
    FloatField,
    FloatSeriesField,
    GitRefField,
    ImageSeriesField,
    IntField,
    LeaderboardEntry,
    NotebookRefField,
    ObjectStateField,
    StringField,
    StringSeriesField,
    StringSetField,
)

if TYPE_CHECKING:
    from neptune.table import Table

PANDAS_AVAILABLE_TYPES = Union[str, float, int, bool, datetime, None]


class FieldToPandasValueVisitor(FieldVisitor[PANDAS_AVAILABLE_TYPES]):

    def visit_float(self, field: FloatField) -> float:
        return field.value

    def visit_int(self, field: IntField) -> int:
        return field.value

    def visit_bool(self, field: BoolField) -> bool:
        return field.value

    def visit_string(self, field: StringField) -> str:
        return field.value

    def visit_datetime(self, field: DateTimeField) -> datetime:
        return field.value

    def visit_file(self, field: FileField) -> None:
        return None

    def visit_string_set(self, field: StringSetField) -> Optional[str]:
        return ",".join(field.values)

    def visit_float_series(self, field: FloatSeriesField) -> Optional[float]:
        return field.last

    def visit_string_series(self, field: StringSeriesField) -> Optional[str]:
        return field.last

    def visit_image_series(self, field: ImageSeriesField) -> None:
        return None

    def visit_file_set(self, field: FileSetField) -> None:
        return None

    def visit_git_ref(self, field: GitRefField) -> Optional[str]:
        return field.commit.commit_id if field.commit is not None else None

    def visit_object_state(self, field: ObjectStateField) -> str:
        return field.value

    def visit_notebook_ref(self, field: NotebookRefField) -> Optional[str]:
        return field.notebook_name

    def visit_artifact(self, field: ArtifactField) -> str:
        return field.hash


def make_row(entry: LeaderboardEntry, to_value_visitor: FieldVisitor) -> Dict[str, PANDAS_AVAILABLE_TYPES]:
    row: Dict[str, PANDAS_AVAILABLE_TYPES] = dict()

    for field in entry.fields:
        value = to_value_visitor.visit(field)
        if value is not None:
            row[field.path] = value

    return row


def sort_key(field: str) -> Tuple[int, str]:
    domain = field.split("/")[0]
    if domain == "sys":
        return 0, field
    if domain == "monitoring":
        return 2, field
    return 1, field


def to_pandas(table: Table) -> pd.DataFrame:

    to_value_visitor = FieldToPandasValueVisitor()
    rows = dict((n, make_row(entry, to_value_visitor)) for (n, entry) in enumerate(table._entries))

    df = pd.DataFrame.from_dict(data=rows, orient="index")
    df = df.reindex(sorted(df.columns, key=sort_key), axis="columns")

    return df



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "prepare_nql_query",
]

from typing import (
    Iterable,
    List,
    Optional,
    Union,
)

from neptune.internal.backends.nql import (
    NQLAggregator,
    NQLAttributeOperator,
    NQLAttributeType,
    NQLQuery,
    NQLQueryAggregate,
    NQLQueryAttribute,
    RawNQLQuery,
)
from neptune.internal.utils.run_state import RunState


def prepare_nql_query(
    ids: Optional[Iterable[str]],
    states: Optional[Iterable[str]],
    owners: Optional[Iterable[str]],
    tags: Optional[Iterable[str]],
    trashed: Optional[bool],
) -> NQLQueryAggregate:
    query_items: List[Union[NQLQueryAttribute, NQLQueryAggregate]] = []

    if trashed is not None:
        query_items.append(
            NQLQueryAttribute(
                name="sys/trashed",
                type=NQLAttributeType.BOOLEAN,
                operator=NQLAttributeOperator.EQUALS,
                value=trashed,
            )
        )

    if ids:
        query_items.append(
            NQLQueryAggregate(
                items=[
                    NQLQueryAttribute(
                        name="sys/id",
                        type=NQLAttributeType.STRING,
                        operator=NQLAttributeOperator.EQUALS,
                        value=api_id,
                    )
                    for api_id in ids
                ],
                aggregator=NQLAggregator.OR,
            )
        )

    if states:
        query_items.append(
            NQLQueryAggregate(
                items=[
                    NQLQueryAttribute(
                        name="sys/state",
                        type=NQLAttributeType.EXPERIMENT_STATE,
                        operator=NQLAttributeOperator.EQUALS,
                        value=RunState.from_string(state).to_api(),
                    )
                    for state in states
                ],
                aggregator=NQLAggregator.OR,
            )
        )

    if owners:
        query_items.append(
            NQLQueryAggregate(
                items=[
                    NQLQueryAttribute(
                        name="sys/owner",
                        type=NQLAttributeType.STRING,
                        operator=NQLAttributeOperator.EQUALS,
                        value=owner,
                    )
                    for owner in owners
                ],
                aggregator=NQLAggregator.OR,
            )
        )

    if tags:
        query_items.append(
            NQLQueryAggregate(
                items=[
                    NQLQueryAttribute(
                        name="sys/tags",
                        type=NQLAttributeType.STRING_SET,
                        operator=NQLAttributeOperator.CONTAINS,
                        value=tag,
                    )
                    for tag in tags
                ],
                aggregator=NQLAggregator.AND,
            )
        )

    query = NQLQueryAggregate(items=query_items, aggregator=NQLAggregator.AND)
    return query


def build_raw_query(query: str, trashed: Optional[bool]) -> NQLQuery:
    raw_nql = RawNQLQuery(query)

    if trashed is None:
        return raw_nql

    nql = NQLQueryAggregate(
        items=[
            raw_nql,
            NQLQueryAttribute(
                name="sys/trashed", type=NQLAttributeType.BOOLEAN, operator=NQLAttributeOperator.EQUALS, value=trashed
            ),
        ],
        aggregator=NQLAggregator.AND,
    )
    return nql



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["StructureVersion"]

from enum import Enum


class StructureVersion(Enum):
    # -------------------------------------------------
    # .neptune/
    #     async/
    #         <uuid>/
    #            exec-<num><timestamp>/
    #               container_type
    #               data-1.log
    #               ...
    # -------------------------------------------------
    LEGACY = 1

    # -------------------------------------------------
    # .neptune/
    #     async/
    #         run__<uuid>/
    #            exec-<timestamp>-<date>-<pid>/
    #               data-1.log
    #               ...
    # -------------------------------------------------
    CHILD_EXECUTION_DIRECTORIES = 2

    # -------------------------------------------------
    # .neptune/
    #     async/
    #         run__<uuid>__<pid>__<random_key>/
    #             data-1.log
    #             ...
    # -------------------------------------------------
    DIRECT_DIRECTORY = 3



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from pathlib import Path
from types import TracebackType
from typing import (
    Optional,
    Tuple,
    Type,
)


class AutoCloseable(ABC):
    def __enter__(self) -> AutoCloseable:
        return self

    @abstractmethod
    def close(self) -> None: ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()


class Resource(AutoCloseable):
    @abstractmethod
    def cleanup(self) -> None: ...

    def flush(self) -> None:
        pass

    def close(self) -> None:
        self.flush()

    @property
    @abstractmethod
    def data_path(self) -> Path: ...


class WithResources(Resource):
    @property
    @abstractmethod
    def resources(self) -> Tuple[Resource, ...]: ...

    def flush(self) -> None:
        for resource in self.resources:
            resource.flush()

    def close(self) -> None:
        for resource in self.resources:
            resource.close()

    def cleanup(self) -> None:
        for resource in self.resources:
            resource.cleanup()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_single_page", "iter_over_pages"]

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    Optional,
)

from bravado.client import construct_request  # type: ignore
from bravado.config import RequestConfig  # type: ignore
from bravado.exception import HTTPBadRequest  # type: ignore
from typing_extensions import (
    Literal,
    TypeAlias,
)

from neptune.api.field_visitor import FieldToValueVisitor
from neptune.api.models import (
    Field,
    FieldType,
    LeaderboardEntriesSearchResult,
    LeaderboardEntry,
)
from neptune.api.proto.neptune_pb.api.model.leaderboard_entries_pb2 import ProtoLeaderboardEntriesSearchResultDTO
from neptune.exceptions import NeptuneInvalidQueryException
from neptune.internal.backends.hosted_client import DEFAULT_REQUEST_KWARGS
from neptune.internal.backends.nql import (
    NQLAggregator,
    NQLAttributeOperator,
    NQLAttributeType,
    NQLEmptyQuery,
    NQLQuery,
    NQLQueryAggregate,
    NQLQueryAttribute,
)
from neptune.internal.backends.utils import construct_progress_bar
from neptune.internal.init.parameters import MAX_SERVER_OFFSET
from neptune.typing import ProgressBarType

if TYPE_CHECKING:
    from neptune.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
    from neptune.internal.id_formats import UniqueId


SUPPORTED_ATTRIBUTE_TYPES = {item.value for item in FieldType}

SORT_BY_COLUMN_TYPE: TypeAlias = Literal["string", "datetime", "integer", "boolean", "float"]


class NoLimit(int):
    def __gt__(self, other: Any) -> bool:
        return True

    def __lt__(self, other: Any) -> bool:
        return False

    def __ge__(self, other: Any) -> bool:
        return True

    def __le__(self, other: Any) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        return False

    def __ne__(self, other: Any) -> bool:
        return True


def get_single_page(
    *,
    client: "SwaggerClientWrapper",
    project_id: "UniqueId",
    attributes_filter: Dict[str, Any],
    limit: int,
    offset: int,
    sort_by: str,
    sort_by_column_type: SORT_BY_COLUMN_TYPE,
    ascending: bool,
    types: Optional[Iterable[str]],
    query: Optional["NQLQuery"],
    searching_after: Optional[str],
    use_proto: Optional[bool] = None,
) -> LeaderboardEntriesSearchResult:
    normalized_query = query or NQLEmptyQuery()
    sort_by_column_type = sort_by_column_type if sort_by_column_type else FieldType.STRING.value
    if sort_by and searching_after:
        sort_by_as_nql = NQLQueryAttribute(
            name=sort_by,
            type=NQLAttributeType(sort_by_column_type),
            operator=NQLAttributeOperator.GREATER_THAN,
            value=searching_after,
        )

        if not isinstance(normalized_query, NQLEmptyQuery):
            normalized_query = NQLQueryAggregate(items=[normalized_query, sort_by_as_nql], aggregator=NQLAggregator.AND)
        else:
            normalized_query = sort_by_as_nql

    sorting = (
        {
            "sorting": {
                "dir": "ascending" if ascending else "descending",
                "aggregationMode": "none",
                "sortBy": {
                    "name": sort_by,
                    "type": sort_by_column_type if sort_by_column_type else FieldType.STRING.value,
                },
            }
        }
        if sort_by
        else {}
    )

    params = {
        "projectIdentifier": project_id,
        "type": types,
        "params": {
            **sorting,
            **attributes_filter,
            "query": {"query": str(normalized_query)},
            "pagination": {"limit": limit, "offset": offset},
        },
    }

    try:
        if use_proto:
            result = client.api.searchLeaderboardEntriesProto(**params).response().result
            proto_data = ProtoLeaderboardEntriesSearchResultDTO.FromString(result)
            return LeaderboardEntriesSearchResult.from_proto(proto_data)
        else:
            request_options = DEFAULT_REQUEST_KWARGS.get("_request_options", {})
            request_config = RequestConfig(request_options, True)
            request_params = construct_request(client.api.searchLeaderboardEntries, request_options, **params)

            http_client = client.swagger_spec.http_client

            json_data = (
                http_client.request(request_params, operation=None, request_config=request_config)
                .response()
                .incoming_response.json()
            )

            return LeaderboardEntriesSearchResult.from_dict(json_data)
    except HTTPBadRequest as e:
        title = e.response.json().get("title")
        if title == "Syntax error":
            raise NeptuneInvalidQueryException(nql_query=str(normalized_query))
        raise e


def find_attribute(*, entry: LeaderboardEntry, path: str) -> Optional[Field]:
    return next((attr for attr in entry.fields if attr.path == path), None)


def iter_over_pages(
    *,
    step_size: int,
    limit: Optional[int],
    sort_by: str,
    sort_by_column_type: SORT_BY_COLUMN_TYPE,
    ascending: bool,
    progress_bar: Optional[ProgressBarType],
    max_offset: int = MAX_SERVER_OFFSET,
    **kwargs: Any,
) -> Generator[Any, None, None]:
    searching_after = None
    last_page = None

    data = get_single_page(
        limit=0,
        offset=0,
        sort_by=sort_by,
        ascending=ascending,
        sort_by_column_type=sort_by_column_type,
        searching_after=None,
        **kwargs,
    )
    total = data.matching_item_count

    limit = limit if limit is not None else NoLimit()

    total = total if total < limit else limit

    progress_bar = False if total <= step_size else progress_bar  # disable progress bar if only one page is fetched

    extracted_records = 0

    field_to_value_visitor = FieldToValueVisitor()

    with construct_progress_bar(progress_bar, "Fetching table...") as bar:
        # beginning of the first page
        bar.update(
            by=0,
            total=total,
        )

        while True:
            if last_page:
                searching_after_field = find_attribute(entry=last_page[-1], path=sort_by)
                if not searching_after_field:
                    raise ValueError(f"Cannot find attribute {sort_by} in last page")
                searching_after = field_to_value_visitor.visit(searching_after_field)

            for offset in range(0, max_offset, step_size):
                local_limit = min(step_size, max_offset - offset)
                if extracted_records + local_limit > limit:
                    local_limit = limit - extracted_records

                result = get_single_page(
                    limit=local_limit,
                    offset=offset,
                    sort_by=sort_by,
                    sort_by_column_type=sort_by_column_type,
                    searching_after=searching_after,
                    ascending=ascending,
                    **kwargs,
                )

                # fetch the item count everytime a new page is started (except for the very fist page)
                if offset == 0 and last_page is not None:
                    total += result.matching_item_count

                total = min(total, limit)

                page = result.entries
                extracted_records += len(page)
                bar.update(by=len(page), total=total)

                if not page:
                    return

                yield from page

                if extracted_records == limit:
                    return

                last_page = page



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "AttributeUploadConfiguration",
    "UploadEntry",
    "normalize_file_name",
    "scan_unique_upload_entries",
    "split_upload_files",
    "FileChunk",
    "FileChunker",
    "compress_to_tar_gz_in_memory",
]

from neptune.internal.storage.datastream import (
    FileChunk,
    FileChunker,
    compress_to_tar_gz_in_memory,
)
from neptune.internal.storage.storage_utils import (
    AttributeUploadConfiguration,
    UploadEntry,
    normalize_file_name,
    scan_unique_upload_entries,
    split_upload_files,
)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "DependencyTrackingStrategy",
    "InferDependenciesStrategy",
    "FileDependenciesStrategy",
]

import os
import sys
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    Union,
)

from neptune.internal.utils.logger import get_logger

if sys.version_info >= (3, 8):
    from importlib.metadata import (
        Distribution,
        distributions,
    )
else:
    from importlib_metadata import Distribution, distributions

from neptune.types import File

if TYPE_CHECKING:
    from neptune import Run

logger = get_logger()


class DependencyTrackingStrategy(ABC):
    @abstractmethod
    def log_dependencies(self, run: "Run") -> None: ...


class InferDependenciesStrategy(DependencyTrackingStrategy):
    def log_dependencies(self, run: "Run") -> None:
        dependencies = []

        def sorting_key_func(d: Distribution) -> str:
            _name = d.metadata["Name"]
            return _name.lower() if isinstance(_name, str) else ""

        dists = sorted(distributions(), key=sorting_key_func)

        for dist in dists:
            if dist.metadata["Name"]:
                dependencies.append(f'{dist.metadata["Name"]}=={dist.metadata["Version"]}')

        dependencies_str = "\n".join(dependencies)

        if dependencies_str:
            run["source_code/requirements"].upload(File.from_content(dependencies_str))


class FileDependenciesStrategy(DependencyTrackingStrategy):
    def __init__(self, path: Union[str, os.PathLike]):
        self._path = path

    def log_dependencies(self, run: "Run") -> None:
        if os.path.isfile(self._path):
            run["source_code/requirements"].upload(self._path)
        else:
            logger.error("[ERROR] File '%s' does not exist - skipping dependency file upload.", self._path)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["StatusRunner"]

import sys
import textwrap
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Sequence,
)

from neptune.cli.collect import collect_containers
from neptune.cli.containers import (
    AsyncContainer,
    OfflineContainer,
)
from neptune.cli.utils import get_qualified_name
from neptune.constants import OFFLINE_NAME_PREFIX
from neptune.envs import PROJECT_ENV_NAME
from neptune.internal.backends.api_model import ApiExperiment
from neptune.internal.utils.logger import get_logger

logger = get_logger(with_prefix=False)

if TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend


offline_run_explainer = """
Runs that execute offline are not created on the server and are not assigned to projects;
instead, they are identified by UUIDs like the ones above.
When synchronizing offline runs, specify the workspace and project using the "--project"
flag. Alternatively, you can set the environment variable
{} to the target workspace/project. See the examples below.
""".format(
    PROJECT_ENV_NAME
)


class StatusRunner:
    @staticmethod
    def status(*, backend: "NeptuneBackend", path: Path) -> None:
        containers = collect_containers(path=path, backend=backend)

        if len(containers.not_found_containers) > 0:
            logger.warning(
                "\nWARNING: %s objects was skipped because they do not exist anymore.",
                len(containers.not_found_containers),
            )
        if not any([containers.synced_containers, containers.unsynced_containers, containers.offline_containers]):
            logger.info("There are no Neptune objects in %s", path)
            sys.exit(1)

        StatusRunner.log_unsync_objects(unsynced_containers=containers.unsynced_containers)
        StatusRunner.log_offline_objects(offline_containers=containers.offline_containers)

        if not containers.unsynced_containers:
            logger.info("\nThere are no unsynchronized objects in %s", path)

        if not containers.synced_containers:
            logger.info("\nThere are no synchronized objects in %s", path)

        logger.info("\nPlease run with the `neptune sync --help` to see example commands.")

    @staticmethod
    def log_offline_objects(*, offline_containers: Sequence["OfflineContainer"], info: bool = True) -> None:
        if offline_containers:
            logger.info("Unsynchronized offline objects:")
            for container in offline_containers:
                logger.info("- %s", f"{OFFLINE_NAME_PREFIX}{container.container_id}")

            if info:
                logger.info("\n%s", textwrap.fill(offline_run_explainer, width=90))

    @staticmethod
    def log_unsync_objects(*, unsynced_containers: Sequence["AsyncContainer"]) -> None:
        if unsynced_containers:
            logger.info("Unsynchronized objects:")
            for container in unsynced_containers:
                experiment = container.experiment

                assert experiment is not None  # mypy fix as experiment is present for async containers

                logger.info("- %s%s", get_qualified_name(experiment), trashed(experiment))


def trashed(cont: ApiExperiment) -> str:
    return " (Trashed)" if cont.trashed else ""



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_logger", "get_disabled_logger", "NEPTUNE_LOGGER_NAME"]

import logging
import sys

NEPTUNE_LOGGER_NAME = "neptune"
NEPTUNE_NO_PREFIX_LOGGER_NAME = "neptune_no_prefix"
NEPTUNE_NOOP_LOGGER_NAME = "neptune_noop"
LOG_FORMAT = "[%(name)s] [%(levelname)s] %(message)s"
NO_PREFIX_FORMAT = "%(message)s"


class CustomFormatter(logging.Formatter):
    def format(self, record):
        record.levelname = record.levelname.lower().ljust(len("warning"))
        formatter = logging.Formatter(LOG_FORMAT)
        return formatter.format(record)


class GrabbableStdoutHandler(logging.StreamHandler):
    """
    This class is like a StreamHandler using sys.stdout, but always uses
    whatever sys.stdout is currently set to rather than the value of
    sys.stderr at handler construction time.
    This enables Neptune to capture stdout regardless
    of logging configuration time.
    Based on logging._StderrHandler from standard library.
    """

    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level)

    @property
    def stream(self):
        return sys.stdout


def get_logger(with_prefix: bool = True) -> logging.Logger:
    name = NEPTUNE_LOGGER_NAME if with_prefix else NEPTUNE_NO_PREFIX_LOGGER_NAME

    return logging.getLogger(name)


def get_disabled_logger() -> logging.Logger:
    return logging.getLogger(NEPTUNE_NOOP_LOGGER_NAME)


def _set_up_logging():
    # setup neptune logger so that logging.getLogger(NEPTUNE_LOGGER_NAME)
    # returns configured logger
    neptune_logger = logging.getLogger(NEPTUNE_LOGGER_NAME)
    neptune_logger.propagate = False

    stdout_handler = GrabbableStdoutHandler()
    stdout_handler.setFormatter(CustomFormatter())
    neptune_logger.addHandler(stdout_handler)

    neptune_logger.setLevel(logging.INFO)


def _set_up_no_prefix_logging():
    neptune_logger = logging.getLogger(NEPTUNE_NO_PREFIX_LOGGER_NAME)
    neptune_logger.propagate = False

    stdout_handler = GrabbableStdoutHandler()
    stdout_handler.setFormatter(logging.Formatter(NO_PREFIX_FORMAT))
    neptune_logger.addHandler(stdout_handler)

    neptune_logger.setLevel(logging.INFO)


def _set_up_disabled_logging():
    neptune_logger = logging.getLogger(NEPTUNE_NOOP_LOGGER_NAME)

    neptune_logger.setLevel(logging.CRITICAL)


_set_up_logging()
_set_up_no_prefix_logging()
_set_up_disabled_logging()



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["MetadataFile"]

import os
from json import (
    JSONDecodeError,
    dump,
    load,
)
from pathlib import Path
from typing import (
    Any,
    Dict,
    Optional,
)

from neptune.core.components.abstract import Resource

METADATA_FILE: str = "metadata.json"


class MetadataFile(Resource):
    def __init__(self, data_path: Path, metadata: Optional[Dict[str, Any]] = None):
        self._data_path = data_path
        self._metadata_path: Path = (data_path / METADATA_FILE).resolve(strict=False)
        self._data: Dict[str, Any] = self._read_or_default()

        if metadata:
            for key, value in metadata.items():
                self.__setitem__(key, value)
            self.flush()

    @property
    def data_path(self) -> Path:
        return self._data_path

    def __getitem__(self, item: str) -> Any:
        return self._data[item]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def flush(self) -> None:
        with open(self._metadata_path, "w") as handler:
            dump(self._data, handler, indent=2)

    def _read_or_default(self) -> Dict[str, Any]:
        if self._metadata_path.exists():
            try:
                with open(self._metadata_path, "r") as handler:
                    data: Dict[str, Any] = load(handler)
                    return data
            except (OSError, JSONDecodeError):
                pass

        return dict()

    def cleanup(self) -> None:
        try:
            os.remove(self._metadata_path)
        except OSError:
            pass



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

__all__ = ("LazyOperationProcessorWrapper",)

from pathlib import Path
from typing import (
    Any,
    Callable,
    Optional,
    TypeVar,
)

from neptune.core.components.abstract import Resource
from neptune.core.components.operation_storage import OperationStorage
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.operation_processor import OperationProcessor

RT = TypeVar("RT")


def trigger_evaluation(method: Callable[..., RT]) -> Callable[..., RT]:
    def _wrapper(self: LazyOperationProcessorWrapper, *args: Any, **kwargs: Any) -> RT:
        self.evaluate()
        return method(self, *args, **kwargs)

    return _wrapper


def noop_if_not_evaluated(method: Callable[..., RT]) -> Callable[..., Optional[RT]]:
    def _wrapper(self: LazyOperationProcessorWrapper, *args: Any, **kwargs: Any) -> Optional[RT]:
        if self.is_evaluated:
            return method(self, *args, **kwargs)
        return None

    return _wrapper


def noop_if_evaluated(method: Callable[..., RT]) -> Callable[..., Optional[RT]]:
    def _wrapper(self: LazyOperationProcessorWrapper, *args: Any, **kwargs: Any) -> Optional[RT]:
        if not self.is_evaluated:
            return method(self, *args, **kwargs)
        return None

    return _wrapper


class LazyOperationProcessorWrapper(OperationProcessor):
    def __init__(
        self,
        operation_processor_getter: Callable[[], OperationProcessor],
        post_trigger_side_effect: Optional[Callable[[], Any]] = None,
    ):
        self._operation_processor_getter = operation_processor_getter
        self._post_trigger_side_effect = post_trigger_side_effect
        self._operation_processor: OperationProcessor = None  # type: ignore

    @noop_if_evaluated
    def evaluate(self) -> None:
        self._operation_processor = self._operation_processor_getter()
        self._operation_processor.start()

    @property
    def is_evaluated(self) -> bool:
        return self._operation_processor is not None

    @trigger_evaluation
    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        self._operation_processor.enqueue_operation(op, wait=wait)

    @property
    @trigger_evaluation
    def operation_storage(self) -> OperationStorage:
        return self._operation_processor.operation_storage

    @property
    @trigger_evaluation
    def data_path(self) -> Path:
        if isinstance(self._operation_processor, Resource):
            return self._operation_processor.data_path
        else:
            raise NotImplementedError

    @trigger_evaluation
    def start(self) -> None:
        self._operation_processor.start()

    @noop_if_not_evaluated
    def pause(self) -> None:
        self._operation_processor.pause()

    @noop_if_not_evaluated
    def resume(self) -> None:
        self._operation_processor.resume()

    @noop_if_not_evaluated
    def flush(self) -> None:
        self._operation_processor.flush()

    @noop_if_not_evaluated
    def wait(self) -> None:
        self._operation_processor.wait()

    @noop_if_not_evaluated
    def stop(self, seconds: Optional[float] = None) -> None:
        self._operation_processor.stop(seconds=seconds)

    @noop_if_not_evaluated
    def close(self) -> None:
        self._operation_processor.close()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ProtoLeaderboardEntriesSearchResultDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MATCHING_ITEM_COUNT_FIELD_NUMBER: builtins.int
    TOTAL_GROUP_COUNT_FIELD_NUMBER: builtins.int
    ENTRIES_FIELD_NUMBER: builtins.int
    matching_item_count: builtins.int
    total_group_count: builtins.int
    @property
    def entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ProtoAttributesDTO]: ...
    def __init__(
        self,
        *,
        matching_item_count: builtins.int = ...,
        total_group_count: builtins.int | None = ...,
        entries: collections.abc.Iterable[global___ProtoAttributesDTO] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_total_group_count", b"_total_group_count", "total_group_count", b"total_group_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_total_group_count", b"_total_group_count", "entries", b"entries", "matching_item_count", b"matching_item_count", "total_group_count", b"total_group_count"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_total_group_count", b"_total_group_count"]) -> typing_extensions.Literal["total_group_count"] | None: ...

global___ProtoLeaderboardEntriesSearchResultDTO = ProtoLeaderboardEntriesSearchResultDTO

@typing_extensions.final
class ProtoAttributesDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXPERIMENT_ID_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    PROJECT_ID_FIELD_NUMBER: builtins.int
    ORGANIZATION_ID_FIELD_NUMBER: builtins.int
    PROJECT_NAME_FIELD_NUMBER: builtins.int
    ORGANIZATION_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTES_FIELD_NUMBER: builtins.int
    experiment_id: builtins.str
    type: builtins.str
    project_id: builtins.str
    organization_id: builtins.str
    project_name: builtins.str
    organization_name: builtins.str
    @property
    def attributes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ProtoAttributeDTO]: ...
    def __init__(
        self,
        *,
        experiment_id: builtins.str = ...,
        type: builtins.str = ...,
        project_id: builtins.str = ...,
        organization_id: builtins.str = ...,
        project_name: builtins.str = ...,
        organization_name: builtins.str = ...,
        attributes: collections.abc.Iterable[global___ProtoAttributeDTO] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "experiment_id", b"experiment_id", "organization_id", b"organization_id", "organization_name", b"organization_name", "project_id", b"project_id", "project_name", b"project_name", "type", b"type"]) -> None: ...

global___ProtoAttributesDTO = ProtoAttributesDTO

@typing_extensions.final
class ProtoAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    INT_PROPERTIES_FIELD_NUMBER: builtins.int
    FLOAT_PROPERTIES_FIELD_NUMBER: builtins.int
    STRING_PROPERTIES_FIELD_NUMBER: builtins.int
    BOOL_PROPERTIES_FIELD_NUMBER: builtins.int
    DATETIME_PROPERTIES_FIELD_NUMBER: builtins.int
    STRING_SET_PROPERTIES_FIELD_NUMBER: builtins.int
    FLOAT_SERIES_PROPERTIES_FIELD_NUMBER: builtins.int
    name: builtins.str
    type: builtins.str
    @property
    def int_properties(self) -> global___ProtoIntAttributeDTO: ...
    @property
    def float_properties(self) -> global___ProtoFloatAttributeDTO: ...
    @property
    def string_properties(self) -> global___ProtoStringAttributeDTO: ...
    @property
    def bool_properties(self) -> global___ProtoBoolAttributeDTO: ...
    @property
    def datetime_properties(self) -> global___ProtoDatetimeAttributeDTO: ...
    @property
    def string_set_properties(self) -> global___ProtoStringSetAttributeDTO: ...
    @property
    def float_series_properties(self) -> global___ProtoFloatSeriesAttributeDTO: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        type: builtins.str = ...,
        int_properties: global___ProtoIntAttributeDTO | None = ...,
        float_properties: global___ProtoFloatAttributeDTO | None = ...,
        string_properties: global___ProtoStringAttributeDTO | None = ...,
        bool_properties: global___ProtoBoolAttributeDTO | None = ...,
        datetime_properties: global___ProtoDatetimeAttributeDTO | None = ...,
        string_set_properties: global___ProtoStringSetAttributeDTO | None = ...,
        float_series_properties: global___ProtoFloatSeriesAttributeDTO | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_bool_properties", b"_bool_properties", "_datetime_properties", b"_datetime_properties", "_float_properties", b"_float_properties", "_float_series_properties", b"_float_series_properties", "_int_properties", b"_int_properties", "_string_properties", b"_string_properties", "_string_set_properties", b"_string_set_properties", "bool_properties", b"bool_properties", "datetime_properties", b"datetime_properties", "float_properties", b"float_properties", "float_series_properties", b"float_series_properties", "int_properties", b"int_properties", "string_properties", b"string_properties", "string_set_properties", b"string_set_properties"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_bool_properties", b"_bool_properties", "_datetime_properties", b"_datetime_properties", "_float_properties", b"_float_properties", "_float_series_properties", b"_float_series_properties", "_int_properties", b"_int_properties", "_string_properties", b"_string_properties", "_string_set_properties", b"_string_set_properties", "bool_properties", b"bool_properties", "datetime_properties", b"datetime_properties", "float_properties", b"float_properties", "float_series_properties", b"float_series_properties", "int_properties", b"int_properties", "name", b"name", "string_properties", b"string_properties", "string_set_properties", b"string_set_properties", "type", b"type"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_bool_properties", b"_bool_properties"]) -> typing_extensions.Literal["bool_properties"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_datetime_properties", b"_datetime_properties"]) -> typing_extensions.Literal["datetime_properties"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_float_properties", b"_float_properties"]) -> typing_extensions.Literal["float_properties"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_float_series_properties", b"_float_series_properties"]) -> typing_extensions.Literal["float_series_properties"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_int_properties", b"_int_properties"]) -> typing_extensions.Literal["int_properties"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_string_properties", b"_string_properties"]) -> typing_extensions.Literal["string_properties"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_string_set_properties", b"_string_set_properties"]) -> typing_extensions.Literal["string_set_properties"] | None: ...

global___ProtoAttributeDTO = ProtoAttributeDTO

@typing_extensions.final
class ProtoIntAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    value: builtins.int
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        value: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "value", b"value"]) -> None: ...

global___ProtoIntAttributeDTO = ProtoIntAttributeDTO

@typing_extensions.final
class ProtoFloatAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    value: builtins.float
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        value: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "value", b"value"]) -> None: ...

global___ProtoFloatAttributeDTO = ProtoFloatAttributeDTO

@typing_extensions.final
class ProtoStringAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    value: builtins.str
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        value: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "value", b"value"]) -> None: ...

global___ProtoStringAttributeDTO = ProtoStringAttributeDTO

@typing_extensions.final
class ProtoBoolAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    value: builtins.bool
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        value: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "value", b"value"]) -> None: ...

global___ProtoBoolAttributeDTO = ProtoBoolAttributeDTO

@typing_extensions.final
class ProtoDatetimeAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    value: builtins.int
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        value: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "value", b"value"]) -> None: ...

global___ProtoDatetimeAttributeDTO = ProtoDatetimeAttributeDTO

@typing_extensions.final
class ProtoStringSetAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    @property
    def value(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        value: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "value", b"value"]) -> None: ...

global___ProtoStringSetAttributeDTO = ProtoStringSetAttributeDTO

@typing_extensions.final
class ProtoFloatSeriesAttributeDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTE_NAME_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TYPE_FIELD_NUMBER: builtins.int
    LAST_STEP_FIELD_NUMBER: builtins.int
    LAST_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    AVERAGE_FIELD_NUMBER: builtins.int
    VARIANCE_FIELD_NUMBER: builtins.int
    CONFIG_FIELD_NUMBER: builtins.int
    attribute_name: builtins.str
    attribute_type: builtins.str
    last_step: builtins.float
    last: builtins.float
    min: builtins.float
    max: builtins.float
    average: builtins.float
    variance: builtins.float
    @property
    def config(self) -> global___ProtoFloatSeriesAttributeConfigDTO: ...
    def __init__(
        self,
        *,
        attribute_name: builtins.str = ...,
        attribute_type: builtins.str = ...,
        last_step: builtins.float | None = ...,
        last: builtins.float | None = ...,
        min: builtins.float | None = ...,
        max: builtins.float | None = ...,
        average: builtins.float | None = ...,
        variance: builtins.float | None = ...,
        config: global___ProtoFloatSeriesAttributeConfigDTO | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_average", b"_average", "_last", b"_last", "_last_step", b"_last_step", "_max", b"_max", "_min", b"_min", "_variance", b"_variance", "average", b"average", "config", b"config", "last", b"last", "last_step", b"last_step", "max", b"max", "min", b"min", "variance", b"variance"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_average", b"_average", "_last", b"_last", "_last_step", b"_last_step", "_max", b"_max", "_min", b"_min", "_variance", b"_variance", "attribute_name", b"attribute_name", "attribute_type", b"attribute_type", "average", b"average", "config", b"config", "last", b"last", "last_step", b"last_step", "max", b"max", "min", b"min", "variance", b"variance"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_average", b"_average"]) -> typing_extensions.Literal["average"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_last", b"_last"]) -> typing_extensions.Literal["last"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_last_step", b"_last_step"]) -> typing_extensions.Literal["last_step"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max", b"_max"]) -> typing_extensions.Literal["max"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_min", b"_min"]) -> typing_extensions.Literal["min"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_variance", b"_variance"]) -> typing_extensions.Literal["variance"] | None: ...

global___ProtoFloatSeriesAttributeDTO = ProtoFloatSeriesAttributeDTO

@typing_extensions.final
class ProtoFloatSeriesAttributeConfigDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MIN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    UNIT_FIELD_NUMBER: builtins.int
    min: builtins.float
    max: builtins.float
    unit: builtins.str
    def __init__(
        self,
        *,
        min: builtins.float | None = ...,
        max: builtins.float | None = ...,
        unit: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_max", b"_max", "_min", b"_min", "_unit", b"_unit", "max", b"max", "min", b"min", "unit", b"unit"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_max", b"_max", "_min", b"_min", "_unit", b"_unit", "max", b"max", "min", b"min", "unit", b"unit"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max", b"_max"]) -> typing_extensions.Literal["max"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_min", b"_min"]) -> typing_extensions.Literal["min"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_unit", b"_unit"]) -> typing_extensions.Literal["unit"] | None: ...

global___ProtoFloatSeriesAttributeConfigDTO = ProtoFloatSeriesAttributeConfigDTO



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "SupportsNamespaces",
    "AbstractNeptuneObject",
    "NeptuneObjectCallback",
    "ProgressBarCallback",
    "ProgressBarType",
]

import abc
import contextlib
from typing import (
    Any,
    Optional,
    Type,
    Union,
)

from typing_extensions import TypeAlias

from neptune.objects.abstract import (
    AbstractNeptuneObject,
    NeptuneObjectCallback,
    SupportsNamespaces,
)


class ProgressBarCallback(contextlib.AbstractContextManager):
    """Abstract base class for progress bar callbacks.

    You can use this class to implement your own progress bar callback that will be invoked in table fetching methods:

    - `fetch_runs_table()`
    - `fetch_models_table()`
    - `fetch_model_versions_table()`

    Example using `click`:
        >>> from typing import Any, Optional, Type
        >>> from types import TracebackType
        >>> from neptune.typing import ProgressBarCallback
        >>> class ClickProgressBar(ProgressBarCallback):
        ...     def __init__(self, *, description: Optional[str] = None, **_: Any) -> None:
        ...         super().__init__()
        ...         from click import progressbar
        ...
        ...         self._progress_bar = progressbar(iterable=None, length=1, label=description)
        ...
        ...     def update(self, *, by: int, total: Optional[int] = None) -> None:
        ...         if total:
        ...             self._progress_bar.length = total
        ...         self._progress_bar.update(by)
        ...
        ...     def __enter__(self) -> "ClickProgressBar":
        ...         self._progress_bar.__enter__()
        ...         return self
        ...
        ...     def __exit__(
        ...         self,
        ...         exc_type: Optional[Type[BaseException]],
        ...         exc_val: Optional[BaseException],
        ...         exc_tb: Optional[TracebackType],
        ...     ) -> None:
        ...         self._progress_bar.__exit__(exc_type, exc_val, exc_tb)
        >>> from neptune import init_project
        >>> with init_project() as project:
        ...     project.fetch_runs_table(progress_bar=ClickProgressBar)
        ...     project.fetch_models_table(progress_bar=ClickProgressBar)

        IMPORTANT: Pass a type, not an instance to the `progress_bar` argument.
        That is, `ClickProgressBar`, not `ClickProgressBar()`.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def update(self, *, by: int, total: Optional[int] = None) -> None: ...


ProgressBarType: TypeAlias = Union[bool, Type[ProgressBarCallback]]
